import pickle
from Bio.Seq import Seq
import os
import pandas as pd
import numpy as np
try:
    from tkinter import filedialog
except:
    pass
import warnings
import glob          

class ManageImportFiles:
    def __init__(self):
        """This class can be used to create the sequencing report from a directory containing tsv files. It was primarily designed for the implementation in Platforma.bio. 
            Use the method merge_tsvs and input the directory
        
        """
        self.columns_not_wanted = ['refPoints', 'allVHitsWithScore', 'allDHitsWithScore',
                        'allJHitsWithScore', 'allCHitsWithScore', 'allVAlignments',
                        'allDAlignments', 'allJAlignments', 'allCAlignments']
        
    @staticmethod    
    def check_cols(single_table: pd.DataFrame, filename:str):
        cols = single_table.columns.to_list()
        nuc_regions = ["nSeqCDR1", "nSeqFR2", "nSeqCDR2", "nSeqFR3", "nSeqCDR3", "nSeqFR4"]
        aa_regions = ["aaSeqFR1", "aaSeqCDR1", "aaSeqFR2", "aaSeqCDR2","aaSeqFR3", "aaSeqCDR3", "aaSeqFR4"]
        mandatory_cols = ["readFraction", "readCount", "cloneId", "targetSequences"]
        assert any(item in nuc_regions for item in cols), "No region starting with nSeq found."
        assert any(item in aa_regions for item in cols), "No region starting with aaSeq found."
        for col in mandatory_cols:
            assert col in cols, f"{col} does not exist in {filename}"
        

    @staticmethod
    def get_filename(tsv_filename:str):
        whole_basename:str = os.path.basename(tsv_filename)
        experiment_name = whole_basename.split(".tsv")[0]
        return experiment_name
    
    def merge_tsvs(self, dir_tsvs:str):
        assert os.path.isdir(dir_tsvs), f"{dir_tsvs} does not exist"
        tsv_files = glob.glob(os.path.join(dir_tsvs, "*.tsv*"))
        sequencing_report = pd.DataFrame([])
        for tsv_table in tsv_files:
            experiment_name = self.get_filename(tsv_table)
            sample = pd.read_table(tsv_table)
            self.check_cols(sample, tsv_table)
            sample.drop(columns=self.columns_not_wanted, inplace=True)
            sample["Experiment"] = experiment_name #you dont need to worry about the length of the name because in Platforma they will decide how the headers look like
            sample["cloneFraction"] = sample["readFraction"] 
            sequencing_report = pd.concat([sequencing_report, sample])
        
        return sequencing_report

dir_tsv = r"src\ExpoSeq\software_tests\test_files\test_show\tables_mixcr"
TSVManager = ManageImportFiles()
seq_report = TSVManager.merge_tsvs(dir_tsv)


class SequencingReport:
    def __init__(self,sequencing_report):
        """This class can be used to tidy and prepare the sequencing report for the pipeline. Alternatively you can input a string as a path to a directory with tsv files to generate a merged object out of these table which can be prepared as sequencing report then.

        Args:
            sequencing_report: _description_

        Returns:
            pd.DataFrame: tidied sequencing report
        """
        if isinstance(sequencing_report, pd.DataFrame) == True: # for pipeline purposes: 
            sequencing_report = sequencing_report.dropna(subset = ["Experiment"])
        elif type(sequencing_report) == str: # fur Platforma purposes to input tsv files
            TSVManager = ManageImportFiles()
            sequencing_report = TSVManager.merge_tsvs(sequencing_report) # contains test for path checking
        else:
            return ValueError
        
        self.origin_seq_report = sequencing_report.copy()
        self.sequencing_report = sequencing_report



    def get_exp_names(self, experiment_path):
        try:
            with open(experiment_path, "rb") as f:
                unique_experiments = pickle.load(f)
        except:
            unique_experiments = self.sequencing_report["Experiment"].unique().tolist()
            unique_experiments = dict(zip(unique_experiments, unique_experiments))
        return unique_experiments
            
    def map_exp_names(self, unique_experiments):
        self.sequencing_report["Experiment"] = self.sequencing_report["Experiment"].map(unique_experiments)


    def is_divisible_by_three(self, seq):
        return len(seq) % 3 == 0

    @staticmethod
    def translate_seq(nucleotide):
        return Seq(nucleotide).translate()

    def get_fragment(self):
        threshold_count = 0.1 * len(self.origin_seq_report) # to get sequences which have values
        avail_cols = [col for col in self.origin_seq_report.columns if col.startswith('nSeq') and self.origin_seq_report[col].count() > threshold_count]
        if "targetSequences" in self.origin_seq_report.columns.tolist():
            avail_cols += ["targetSequences"]
        else:
            pass
        return avail_cols

    
    def filter_region(self, region_string, remove_gaps = True, remove_errors = True):
        fixed_cols = ["Experiment", "cloneId", "readCount", "readFraction"]
        if region_string == "targetSequences":
            aa_string = "aaSeqtargetSequences"
            nseq_string = "nSeqtargetSequences"
            self.sequencing_report = self.origin_seq_report.copy()
            self.sequencing_report.rename(columns={region_string: nseq_string}, inplace=True) 
            if remove_gaps:
                self.sequencing_report = self.sequencing_report[self.sequencing_report['nSeq' + region_string].apply(self.is_divisible_by_three)]
            else:
                warnings.warn("You decided to not remove the gaps which means that sequences potentially are not divisible by 3.\nThere is no functionality for finding the open reading frame. Thus, translating the full sequence to a peptide sequence will start at position 1.")
            self.sequencing_report[aa_string] = self.sequencing_report[nseq_string].apply(self.translate_nucleotide_to_amino_acid) 
            
            added_columns = ["nSeq" + region_string, "aaSeq" + region_string] #"minQual" + region_string,
            cols_of_interest = fixed_cols + added_columns
            self.sequencing_report = self.sequencing_report[cols_of_interest]
        else:
            added_columns = ["nSeq" + region_string, "aaSeq" + region_string] #"minQual" + region_string,
            cols_of_interest = fixed_cols + added_columns
            self.sequencing_report = self.origin_seq_report[cols_of_interest]
            if remove_gaps:
                self.sequencing_report = self.sequencing_report[self.sequencing_report['nSeq' + region_string].apply(self.is_divisible_by_three)] # removes gaps indirectly
        
        if remove_errors:
            self.sequencing_report = self.remove_seq_errors(self.sequencing_report, region_string)

       # self.sequencing_report["aaSeq" + region_string] = self.sequencing_report["nSeq" + region_string].apply(self.translate_sequence)
    
    @staticmethod
    def translate_nucleotide_to_amino_acid(nucleotide_sequence):
        seq = Seq(nucleotide_sequence)
        return str(seq.translate())    

    def trim_data(self, region_string, length_threshold = 9, min_read_count = 0, new_fraction = "cloneFraction"):
        aa_string = "aaSeq" + region_string
        nseq_string = "nSeq" + region_string
        assert aa_string in self.sequencing_report.columns.tolist(), f"{self.sequencing_report.columns.tolist()}"
        assert nseq_string in self.sequencing_report.columns.tolist(), f"{self.sequencing_report.columns.tolist()}"
        length_threshold = length_threshold - 1
        if min_read_count > 0:
            min_read_count = min_read_count - 1
        
        sequencing_report = self.sequencing_report.groupby("Experiment", group_keys = True).apply(lambda group: group.drop_duplicates(subset=[aa_string], keep="first")).reset_index(drop=True)

        indexes_to_drop = sequencing_report[sequencing_report[aa_string] == 'region_not_covered'].index
        sequencing_report = sequencing_report.drop(indexes_to_drop)

        sequencing_report["lengthOfCDR3"] = sequencing_report[nseq_string].str.len() ## assumes all that pipeline starts with cdr3 region

        sequencing_report = sequencing_report.reset_index()

        sequencing_report = sequencing_report[(sequencing_report["aaSeq" + region_string].str.len()) > length_threshold ]
        sequencing_report = sequencing_report[(sequencing_report["readCount"] > (min_read_count + 1))]
        new_column = sequencing_report['readCount'] / sequencing_report.groupby('Experiment')['readCount'].transform('sum')
        self.sequencing_report = sequencing_report.copy()
        self.sequencing_report[new_fraction] = np.array(new_column)
        self.sequencing_report.drop(columns = ["readFraction"], inplace = True)
        self.sequencing_report.drop(columns = ["index"], inplace = True)
        
        
    def remove_not_covered(self):
        self.sequencing_report = self.sequencing_report[~self.sequencing_report.applymap(lambda x: x == "region_not_covered").any(axis=1)]

    
    @staticmethod
    def remove_seq_errors(sequencing_report, region_string):
        sequencing_report = sequencing_report.loc[~sequencing_report["aaSeq" + region_string].str.contains("[*]")]
        return sequencing_report
    
    def prepare_seq_report(self, region_string, length_threshold, min_read_count, remove_gaps = True, remove_errors = True):
        self.filter_region(region_string, remove_gaps, remove_errors) # sequencing errors are removed here and sequences with gaps are indirectly removed with: divisible_by = 3
        self.trim_data(region_string, length_threshold, min_read_count,)
        self.remove_not_covered()
       # self.remove_seq_errors()

    
    def check_sample_name(self, module_dir, experiment_name):
        sample_names = self.origin_seq_report["Experiment"].unique().tolist()
        for sample in sample_names:
            if len(sample) > 22:
                new_sample_name = input(f"Your sample name for {sample} is to long. Please enter a new name with maximum 22 characters.")
                replacement_mapping = {sample: new_sample_name}
                self.origin_seq_report['Experiment'] = self.origin_seq_report['Experiment'].replace(replacement_mapping)
                self.sequencing_report["Experiment"] = self.sequencing_report["Experiment"].replace(replacement_mapping)
        path_file = os.path.join(module_dir, "my_experiments", experiment_name, "sequencing_report.csv")
        if os.path.isdir(os.path.dirname(path_file)):
            self.origin_seq_report.to_csv(path_file, index = False)
        
    def renew_exp_names_origin(self, replacement_mapping, path_report):
        self.origin_seq_report["Experiment"] = self.origin_seq_report["Experiment"].replace(replacement_mapping)
        user_input = input("Do you want to save and change the new names in the sequencing report? - If so, you will not have to change the names again if you restart the program. Y/n")
        while True:
            if user_input.lower() in ["Y", "y", "n", "N"]:
                if user_input.lower() in ["Y", "y"]:
                    self.origin_seq_report.to_csv(os.path.join(path_report, "sequencing_report.csv"), index = False)
                else:
                    pass
                break
            else:
                print("Please enter Y or n")
            
        


class BindingReport:
    def __init__(self, module_dir, experiment):
        self.binding_data_dir = os.path.join(module_dir,
                                "my_experiments",
                                experiment,
                                "binding_data.csv")
        
    @staticmethod
    def collect_binding_data(binding_data = None):
        if binding_data is None:
            binding_data = pd.DataFrame([])
        else:
            pass
        second_prompt = False
        while True:
            # prompt the user to add a file
            if not second_prompt:
                print("You can either add an excel sheet or a csv file which contains the binding data.\nNote: the first column must contain the CDR3 sequences and its column name has to be aaSeqCDR3.")
            else:
                print("Please modify the file and choose it again.")
            try:
                binding_file = filedialog.askopenfilename()
            except:
                while True:
                    binding_file = input("copy and paste the path to your binding report")
                    if os.path.isfile(os.path.abspath(binding_file)):
                        break
                    else:
                        print("Please enter a valid filepath. ")
            
            if binding_file.endswith(".xlsx") or binding_file.endswith(".csv") or binding_file.endswith(".tsv"):
                if binding_file.endswith(".xlsx"):
                    binding_new = pd.read_excel(binding_file)
                elif binding_file.endswith(".csv"):
                    binding_new = pd.read_csv(binding_file)
                elif binding_file.endswith(".tsv"):
                    binding_new = pd.read_table(binding_file)
                if binding_new.columns.to_list()[0] == "aaSeqCDR3":
                    second_prompt = False
                    pass
                else:
                    second_prompt = True
                    print("Please chechk if your file follows the following requirements:\n1. The first column must contain the CDR3 sequences and its column name has to be aaSeqCDR3.\n2. Check that you separate your data by comma or tab.\n3. Your file needs to end with .xlsx, .csv or .tsv")
            else:
                print("Please enter a valid filepath to a csv or xlsx file")
                second_prompt = True

            if not second_prompt:
                binding_data = pd.concat([binding_data, binding_new])
                response = input("Do you want to continue adding files? (Y/n) ")
                if response.lower() == "n":
                    break
                print("The first five rows of your binding data look like this:")
                print(binding_data.head(5))
        return binding_data
        
    def ask_binding_data(self):
        if not os.path.isfile(self.binding_data_dir):
            add_binding = input("Do you have binding Data? Y/n")
            if add_binding.lower() in ["Y", "y"]:
                binding_data = self.collect_binding_data()

                binding_data.to_csv("binding_data.csv")
            else:
                binding_data = None
        else:
            binding_data = pd.read_csv(self.binding_data_dir)
        if binding_data is not None:
            binding_data.drop_duplicates(subset = "aaSeqCDR3", inplace = True)
        return binding_data








import matplotlib.pyplot as plt
import logomaker
import numpy as np
from ..settings.layout_finder import best_layout
from textwrap import wrap
import pandas as pd
import warnings
from iglabel import IMGT

class PrepareData:
    @staticmethod
    def check_args(samples, report, chosen_seq_length,):
        assert type(samples) == list, "You have to give a list as input for the sample"

        if chosen_seq_length != None:
            assert (
                type(chosen_seq_length) == int
            ), "You have to give an integer as input for the sequence length you want to analyze"
        for sample in samples:
            assert (
                sample in report["Experiment"].unique().tolist()
            ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"

        
    @staticmethod
    def calculate_entropy(probs):
        """Calculate Shannon entropy."""
        return -np.sum(
            [p * np.log2(p) if p > 0 else 0 for p in probs]
        )  # https://biology.stackexchange.com/questions/64368/how-to-determine-the-height-bits-in-a-sequence-logo

    @staticmethod
    def get_labels(region_string:str, chosen_seq_length:int):
        """Creates the IMGT labels for the chosen sequence length

        Args:
            region_string (str): _description_
            chosen_seq_length (int): _description_

        Returns:
            list: A list with strings of either the IMGT labels or the position numbers
        """
        if region_string != "targetSequences":
            region = [region_string.replace("aaSeq", "")]
            label_dict, _ = IMGT([chosen_seq_length * "A"], region, save = False)
            numbers_true = list(label_dict.values())[0]
        else:
            numbers_true = [str(i) for i in list(range(1, chosen_seq_length + 1))]
        return numbers_true

    def cleaning(self, samples, report, chosen_seq_length, region_string, method):
        self.check_args(samples, report, chosen_seq_length)
        local_report = report.loc[report["Experiment"].isin(samples)]
       # sample = report[report["Experiment"] == sample_name]
        local_report = local_report[["Experiment", "cloneFraction", region_string]]
        sequences = local_report[region_string]
        if chosen_seq_length != None:
            assert sequences.str.len().max() >= chosen_seq_length, "Your chosen seq length must exist in your sequence repertoire"
        aminoacids = "ACDEFGHIKLMNPQRSTVWY"

        if chosen_seq_length != None:
            sequences = local_report[
                local_report[region_string].astype(str).str.len() == chosen_seq_length
            ][region_string]
        else:
            warnings.warn(
                "You will no have all sequence length included but the problem of that is that you assume that sequences with different lengths have the same properties on relative amino acid positions, which is not necessarily true."
            )
            sequences = local_report[region_string]
            chosen_seq_length = int(sequences.str.len().max())
            
        compDict = {aa: chosen_seq_length * [0] for aa in aminoacids}
        length_filtered_seqs = sequences.shape[0]
        for seq in sequences:
            for aa_position in range(len(seq)):
                aminoacid = seq[aa_position]
                if aminoacid == "*" or aminoacid == "_":
                    pass
                else:
                    compDict[aminoacid][aa_position] += 1
        if method == "bits":
            # Calculate frequencies
            frequencies = {
                aa: [count / length_filtered_seqs for count in compDict[aa]]
                for aa in aminoacids
            }

            # Calculate Shannon entropy for each position
            entropies = [
                self.calculate_entropy([frequencies[aa][i] for aa in aminoacids])
                for i in range(chosen_seq_length)
            ]

            # Calculate bits for sequence logo for each amino acid
            bits_dict = {
                aa: [
                    2 - entropies[i] if frequencies[aa][i] > 0 else 0
                    for i in range(chosen_seq_length)
                ]
                for aa in aminoacids
            }
            aa_distribution = pd.DataFrame.from_dict(bits_dict)
        else:
            aa_distribution = pd.DataFrame.from_dict(compDict)
            aa_distribution = aa_distribution.divide(
                aa_distribution.sum(axis=1), axis=0
            )
            aa_distribution.astype("float16")
        return aa_distribution


class LogoPlot:
    def __init__(
        self,
        ax,
        sequencing_report,
        region_string,
        sample,
        highlight_spec_position,
        font_settings,
        chosen_seq_length,
        method,
        color_scheme,
        avail_regions,
        **kwargs,
    ):
        self.avail_regions = avail_regions
        self.region_string = region_string
        self.ax = ax
        self.chosen_seq_length = self.find_seq_length(
            sequencing_report, sample, chosen_seq_length, region_string
        )
        self.aa_distribution = PrepareData().cleaning(
            sample, sequencing_report, self.chosen_seq_length, region_string, method
        )
        # self.createPlot()
        self.font_settings = font_settings
        self.func_type = {
            "alpha": [0.01, 0.01, 0.01, 1],
            "shade": [0.01, 0.01, 0.01, 1],
            "fade": [0.01, 0.01, 0.01, 1],
        }
        self.button_type = {
            "alpha": "CustomScale",
            "shade": "CustomScale",
            "fade": "CustomScale",
        }
        self.createPlot(color_scheme=color_scheme, **kwargs)
        self.add_style(highlight_spec_position, sample)

    @staticmethod
    def find_seq_length(
        sequencing_report, sample, chosen_seq_length, region_of_interest
    ):
        filtered_data = sequencing_report.loc[sequencing_report["Experiment"].isin(sample)]
        length_counts = filtered_data[region_of_interest].str.len().value_counts()

        if chosen_seq_length == None:
            max_length = length_counts.idxmax()
            chosen_seq_length = max_length
        else:
            if chosen_seq_length in length_counts:
                pass
            else:
                print(
                    "The chosen sequence length is not in the data. The most frequent sequence length was chosen instead"
                )
                max_length = length_counts.idxmax()
                chosen_seq_length = max_length
            
        return int(chosen_seq_length)

    def createPlot(
        self,
        shade_below=0.5,
        fade_below=0.5,
        font_name="Arial Rounded MT Bold",
        color_scheme="skylign_protein",
        show_spines=False,
    ):

        self.logo_plot = logomaker.Logo(
            self.aa_distribution,
            shade_below=shade_below,
            fade_below=fade_below,
            font_name=font_name,
            color_scheme=color_scheme,
            show_spines=show_spines,
            ax=self.ax,
        )
        self.logo_plot.style_xticks(anchor=1, spacing=1, rotation=0)

            
    def add_style(self, highlight_specific_pos, sample):
        if "fontsize" in self.font_settings.keys():
            original_fontsize = self.font_settings["fontsize"]
            self.ax.set_ylabel("Frequency", **self.font_settings)
            self.ax.set_xlabel("Position on sequence", **self.font_settings)
            self.font_settings["fontsize"] = 22
            title = "\n".join(
                wrap(
                    "Logo Plot of "
                    + " ".join(sample)
                    + " with sequence length "
                    + str(self.chosen_seq_length),
                    width=40,
                )
            )
            plt.title(title, **self.font_settings)
            self.font_settings["fontsize"] = original_fontsize
            labels_true = list(range(0, self.chosen_seq_length))
            if self.region_string != "targetSequences":
                region = [self.region_string.replace("aaSeq", "")]
                label_dict, _ = IMGT([self.chosen_seq_length * "A"], region, save = False)
                numbers_true = list(label_dict.values())[0]
            else:
                numbers_true = list(range(1, self.chosen_seq_length + 1))
            assert len(numbers_true) == len(labels_true), f" you have {len(numbers_true)} labels and {len(labels_true)} and xticsk"
            plt.xticks(labels_true, numbers_true)
            if highlight_specific_pos != None:
                self.logo_plot.highlight_position(p=5, color="gold", alpha=0.5)


def plot_logo_multi(
    fig,
    sequencing_report,
    samples,
    font_settings,
    region_string,
    method,
    chosen_seq_length=16,
):
    if samples == "all":
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    else:
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.array([i for i in unique_experiments if i in samples])
        unique_experiments = np.sort(unique_experiments)
    Tot = unique_experiments.shape[0]
    Rows, Cols = best_layout(Tot)
    Position = range(1, Tot + 1)
    n = 0
    adapted_fontsize = 10 - int(Cols) + 2
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = adapted_fontsize
    #  fig = plt.figure(1, constrained_layout=True)
    for i in unique_experiments:
        aa_distribution = PrepareData().cleaning(
            [i], sequencing_report, chosen_seq_length, region_string, method
        )
        if aa_distribution.isna().any().any():
            continue

        ax = fig.add_subplot(
            Rows, Cols, Position[n], xticks=(np.arange(0, chosen_seq_length, step=1))
        )
        logo_plot = logomaker.Logo(
            aa_distribution,
            shade_below=0.5,
            fade_below=0.5,
            font_name="Arial Rounded MT Bold",
            color_scheme="skylign_protein",
            show_spines=False,
            ax=ax,
            allow_nan=True,
        )
        # logo_plot.set_xticks(range(aa_distribution.shape[0]))
        logo_plot.style_xticks(
            anchor=0,
            spacing=1,
            rotation=0,
        )
        plt.title(
            i, **font_settings
        )  # check out https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.title.html

        n = n + 1
    else:
        print("Sample " + i + "was skipped because no sequence was found")
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    fig.suptitle(
        "Logo Plots for sequence Length " + str(chosen_seq_length), **font_settings
    )
    font_settings["fontsize"] = original_fontsize



import torch
import re
import pandas as pd
# + pip install sentencepiece
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
import numpy as np 
import umap 
from sklearn.cluster import DBSCAN

class TransformerBased:

    def __init__(self, choice='Rostlab/prot_t5_xl_half_uniref50-enc', truncation = False, max_length = None):
        self.truncation = truncation
        self.max_length = max_length
        self.choice = choice
        self.prep_model()

    def prep_model(self):
        model_types = ["facebook/esm2_t6_8M_UR50D", "Rostlab/ProstT5_fp16", "Rostlab/prot_t5_xl_uniref50", "Rostlab/prot_t5_base_mt_uniref50", "Rostlab/prot_bert_bfd_membrane", "Rostlab/prot_t5_xxl_uniref50",
                       "Rostlab/ProstT5", "Rostlab/prot_t5_xl_half_uniref50-enc", "Rostlab/prot_bert_bfd_ss3", "Rostlab/prot_bert_bfd_localization",
                       "Rostlab/prot_t5_xl_bfd", "Rostlab/prot_bert", "Rostlab/prot_xlnet", "Rostlab/prot_bert_bfd", "Rostlab/prot_t5_xxl_bfd"]
        
        t5 = ["Rostlab/ProstT5", "Rostlab/ProstT5_fp16", "Rostlab/prot_t5_xl_uniref50", "Rostlab/prot_t5_xxl_uniref50", "Rostlab/prot_t5_xl_half_uniref50-enc", "Rostlab/prot_t5_xl_bfd", "Rostlab/prot_t5_xxl_bfd"]
        bertis = ["Rostlab/prot_bert_bfd_ss3", "Rostlab/prot_bert_bfd_localization", "Rostlab/prot_bert", "Rostlab/prot_xlnet", "Rostlab/prot_bert_bfd"]
        esm = ["facebook/esm2_t6_8M_UR50D"]
        assert self.choice in model_types, f"Your choice has to be a model from https://huggingface.co/Rostlab and has to be one of {model_types}"
        if self.choice in t5:
            from transformers import T5Tokenizer, T5EncoderModel
            self.tokenizer = T5Tokenizer.from_pretrained(self.choice,
                                                    do_lower_case=False,)

            self.model = T5EncoderModel.from_pretrained(self.choice)
        elif self.choice in bertis:
            from transformers import BertModel, BertTokenizer
            self.tokenizer = BertTokenizer.from_pretrained(self.choice, do_lower_case=False )
            self.model = BertModel.from_pretrained(self.choice)
        elif self.choice in esm:
            from transformers import AutoTokenizer, EsmModel
            self.tokenizer = AutoTokenizer.from_pretrained(self.choice)
            self.model = EsmModel.from_pretrained(self.choice)
        

    @staticmethod
    def filter_sequences(sequencing_report:pd.DataFrame, batch_size:int, experiments:list,binding_data,
                         region_of_interest = "aaSeqCDR3", cf_column_name = "cloneFraction", sample_column_name = "Experiment"):
        
        report_batch = sequencing_report.groupby(sample_column_name).head(batch_size)

        selected_rows = report_batch.loc[report_batch[sample_column_name].isin(experiments)]
        
        selected_rows = selected_rows.drop_duplicates(subset=[region_of_interest], keep='first')
        
        assert selected_rows.shape[0] <= len(experiments) * batch_size
        
        max_clone_fraction = selected_rows.groupby(['Experiment', region_of_interest])['cloneFraction'].max()
        max_clone_fraction = max_clone_fraction.reset_index()
        # Merge back to original DataFrame to filter rows
        result_df = selected_rows.merge(max_clone_fraction, on=[region_of_interest, 'Experiment'], suffixes=('', '_max'))

        selected_rows = result_df[result_df['cloneFraction'] == result_df['cloneFraction_max']]

        # Drop the additional column used for comparison
        selected_rows.drop(columns=['cloneFraction_max'], inplace=True)

        
        assert selected_rows.shape[0] <= len(experiments) * batch_size
        if binding_data is not None:
            mix = selected_rows.merge(binding_data, on = region_of_interest, how = "outer")
            selected_rows = mix.fillna(0)
        max_fraction = max(selected_rows[cf_column_name])
        selected_rows.loc[selected_rows[cf_column_name] == 0.0, cf_column_name] = max_fraction
        selected_rows = selected_rows.sort_values(by=[cf_column_name, sample_column_name], ascending=False)
        sequences_filtered = selected_rows[region_of_interest]
        sequences = [" ".join(list(re.sub(r"[UZOB*_]", "X", sequence))) for sequence in sequences_filtered]
         
        return sequences,sequences_filtered, selected_rows
        
    def prepare_sequences(self,sequences, device = "cpu"):
        ids = self.tokenizer.batch_encode_plus(sequences, add_special_tokens=True, truncation = True, max_length = 1430)
        input_ids = torch.tensor(ids['input_ids'])
        attention_mask = torch.tensor(ids['attention_mask'])
        return attention_mask, input_ids
    
    def get_result(self, sequences):        
        if not self.truncation:    
            encoded_input = self.tokenizer(sequences, return_tensors='pt', padding=True)
        else:
            if self.max_length == None:
                max_length = 1440
            else:
                max_length = self.max_length
            encoded_input = self.tokenizer(sequences, return_tensors = "pt", truncation = True, max_length = max_length)
        embedding_repr = self.model(**encoded_input)
        return embedding_repr
    
    
    def embedding_per_seq(self, sequences, normalize = False):
        sequences_list = []
        for seq in sequences:
            embeddings = self.get_result(seq)
            last_hidden_state = embeddings.last_hidden_state
            maximum_length = last_hidden_state.shape[1]
           
            avg_seq = np.squeeze(last_hidden_state, axis=0)
            
            avg_seq = last_hidden_state.mean(dim = 1) # take average for each feature from all amino acids
            if normalize == True:
                #avg_seq = (avg_seq - torch.min(avg_seq)) / (torch.max(avg_seq) - torch.min(avg_seq))
                min_value = torch.min(avg_seq)
                shifted_distribution = avg_seq - min_value
                sum_shifted = torch.sum(shifted_distribution)
                normalized_distribution = shifted_distribution / sum_shifted
                avg_seq = normalized_distribution
            sequences_list.append(avg_seq.cpu().detach().numpy()[0])
        
        sequences_list = np.array(sequences_list)
        return sequences_list
    
    @staticmethod
    def do_pca(sequences_list, pca_components):
        """
        output: X: principal components of the embedding which has the shape: x = batch_size, y = principal component
        """
      #  assert batch_size > pca_components, "Your batch size (Number of sequences) needs to be bigger than the pca components. Further you should decrease the pca components more, otherwise the dimension reduction does not make really sense."
        # batch size, sequence length, features

        # loop over each sequence and make two dimensional and take average for all amino acids: output size: (batch_size, 1024)
       # sequences_list = self.embedding_per_seq(sequences)
         # shape y = 1024, x = batch_size (No. sequences)
        pca = PCA(n_components=pca_components)
        pca.fit(sequences_list)
        X = pca.transform(sequences_list)
        
        print("Explained variance after reducing to " + str(pca_components) + " dimensions:" + str(np.sum(pca.explained_variance_ratio_).tolist()))

        return X
    
    @staticmethod
    def do_tsne(X, perplexity, iterations_tsne):
        """
        input: vector with shape: x = n_samples and y: n_features
        """
        tsne = TSNE(n_components=2,
            verbose=0,
            perplexity=perplexity,
            n_iter=iterations_tsne)
        tsne_results = tsne.fit_transform(X)
        tsne_results = pd.DataFrame(tsne_results,
                                    columns = ["tsne1", "tsne2"])
        return tsne_results
    @staticmethod
    def do_umap(X, n_neighbors = 15,min_dist = 0.2, random_seed = 42, densmap = True, n_components = 2, y = None, metric = "euclidian", number_jobs = -1, n_epochs = 1000):
        """_summary_

        Args:
            X (array): an array with the PC's from PCA
            n_neighbors (int): Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should be between 5 to 50.
            random_seed: Set a certrain seed for reproducing your results
            densmap: This parameter allows you to visualize points more densily which are also more dense in all dimensions to each other. You can have an idea about this here: https://umap-learn.readthedocs.io/en/latest/densmap_demo.html

        Returns:
            pd.DataFrame: return a pandas dataframe with the two umap_dimensions
        """
        reducer = umap.UMAP(random_state = random_seed, n_neighbors = n_neighbors, 
                            densmap = densmap, n_components = n_components,
                            min_dist = min_dist, metric= metric, n_jobs = number_jobs,
                            n_epochs=n_epochs)
        if y == None:
            reduced_dim = reducer.fit_transform(X)
        else:
            reduced_dim = reducer.fit_transform(X, y = y)
        if reduced_dim.shape[1] == 2:
            results = pd.DataFrame(reduced_dim, columns = ["UMAP_1", "UMAP_2"])
        else: # for multidimensional UMAP
            results = pd.DataFrame([])
            for i in range(reduced_dim.shape[1]):
                results["UMAP_" + str(i+1)] = reduced_dim[:,i]
        return results, reduced_dim
    
    @staticmethod
    def cluster_with_hdbscan(results, eps = 0.5, min_pts = 2):
        """_summary_

        Args:
            results (pd.DataFrame): is the output from do_umap
            eps (float, optional): Maximum distance between two points to still form one cluster. Defaults to 3.
            min_pts (int, optional): fewest number of points required to form a cluster. Defaults to 4.

        Returns:
            _type_: _description_
        """
        arr = results.values
        assert arr.shape[1]==2
        get_clusters = DBSCAN(eps = eps, min_samples = min_pts).fit_predict(arr)
        return get_clusters


from textwrap import wrap
import pandas as pd


class StackedAADistribution:
    def __init__(
        self,
        sequencing_report,
        sample,
        region,
        region_of_interest,
        protein=True,
        font_settings={},
        ax=None,
    ):
        aa_distribution = self.prepare_data(
            sequencing_report, sample, region, region_of_interest
        )

        self.plot(aa_distribution, ax, sample, region, protein, font_settings)

    @staticmethod
    def prepare_data(sequencing_report, sample, region, region_string, chosen_seq_length= None):
        sample = sequencing_report[sequencing_report["Experiment"] == sample]
        local_report = sample[["Experiment", "cloneFraction", region_string]]
        aminoacids = "ACDEFGHIKLMNPQRSTVWY"

        sequences = local_report[
            local_report[region_string].astype(str).str.len() >= region[1]
        ][region_string]
        max_length = local_report[region_string].str.len().max()
        if not region[1] <= max_length:
            raise ValueError(
                "you upper region limit is above the longest sequence. That is not possible. Please reduce it."
            )
        compDict = {aa: max_length * [0] for aa in aminoacids}
        for seq in sequences:
            for aa_position in range(len(seq)):
                aminoacid = seq[aa_position]
                if aminoacid == "*"  or aminoacid == "_":
                    pass
                else:
                    compDict[aminoacid][aa_position] += 1
        aa_distribution = pd.DataFrame.from_dict(compDict)
        aa_distribution = aa_distribution.divide(aa_distribution.sum(axis=1), axis=0)
        aa_distribution = aa_distribution[
            (aa_distribution.index >= region[0]) & (aa_distribution.index <= region[1])
        ]
        return aa_distribution

    @staticmethod
    def get_colors(aa_distribution):
        color_scheme = {
            "F": [0.16, 0.99, 0.18],
            "Y": [0.04, 0.40, 0.05],
            "L": [0.99, 0.60, 0.25],
            "V": [1.0, 0.80, 0.27],
            "I": [0.80, 0.60, 0.24],
            "H": [0.40, 0.02, 0.20],
            "W": [0.42, 0.79, 0.42],
            "A": [0.99, 0.60, 0.42],
            "S": [0.04, 0.14, 0.98],
            "T": [0.17, 1.0, 1.0],
            "M": [0.80, 0.60, 0.80],
            "N": [0.21, 0.40, 0.40],
            "Q": [0.40, 0.41, 0.79],
            "R": [0.59, 0.02, 0.04],
            "K": [0.40, 0.20, 0.03],
            "E": [0.79, 0.04, 0.22],
            "G": [0.95, 0.94, 0.22],
            "D": [0.99, 0.05, 0.11],
            "P": [0.10, 0.61, 0.99],
            "C": [0.09, 0.60, 0.60],
        }
        color_list = [color_scheme[aa] for aa in aa_distribution.columns]
        return color_list

    def plot(self, aa_distribution, ax, sample, region, protein, font_settings):
        color_list = self.get_colors(aa_distribution)
        aa_distribution.plot(kind="bar", stacked=True, color=color_list, ax=ax)
        ax.set_xlabel("Position on amino acid sequence", **font_settings)
        if protein == True:
            ax.set_ylabel("Relatvie Proportion of Amino Acid", **font_settings)
        else:
            ax.set_ylabel("Relatvie Proportion of Nucleotide", **font_settings)

        #    ax.set_xticks(rotation = 360)
        ax.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
        original_fontsize = font_settings["fontsize"]
        font_settings["fontsize"] = 22
        if protein == True:
            title = "\n".join(wrap("Stacked Amino Acid Distribution of " + sample, 40))
            ax.set_title(title, pad=12, **font_settings)
        else:
            title = "\n".join(wrap("Stacked Nucleotide Distribution of " + sample, 40))
            ax.set_title(title, pad=12, **font_settings)
        ax.set_xticklabels([*range(region[0], region[1] + 1)], rotation=0, ha="center")
        font_settings["fontsize"] = original_fontsize




def cluster_single_AG(fig, sequencing_report, antigen, binding_data, max_ld, min_ld, batch_size, region_string,
                      preferred_cmap="Blues", specific_experiments=False, label_type="numbers"):
    assert antigen in binding_data.columns, "it seems like your antigen does not exist in the binding data. Please enter the correct value"
    if specific_experiments != False:
        report_batch = sequencing_report[sequencing_report['Experiment'].isin(specific_experiments)]
    else:
        report_batch = sequencing_report
    report_batch = report_batch.groupby("Experiment").head(batch_size)
    experiments = report_batch["Experiment"].unique()
    Tot = experiments.shape[0]
    Rows, Cols = best_layout(Tot)
    Position = range(1, Tot + 1)
    plot_number = 0
    cluster_report = pd.DataFrame([])
    for experiment in experiments:
        ax = fig.add_subplot(Rows, Cols, Position[plot_number])
        batch = report_batch[report_batch["Experiment"] == experiment]
        batch = batch.dropna(subset = ["aaSeqCDR3"])
        binding_data = binding_data[[antigen, "aaSeqCDR3"]]
        binding_data = binding_data.dropna(subset = ["aaSeqCDR3"])
        mix = pd.merge(batch, binding_data, how = "left", on = "aaSeqCDR3")
        mix = mix.fillna(0)
        mix = mix.reset_index()
        G = cleaning(mix, max_ld, min_ld, region_string)
        # remove nodes that are not connected to any other nodes
        G_deg = G.degree()
        to_remove = [n for (n, deg) in G_deg if deg == 0]
        G.remove_nodes_from(to_remove)
       #fr print(mix.)
        nodesize = []
        clone_counts = mix["readCount"]
        color_values = []
        label_numbers = {}
        label_sequences = {}
        for index, g in enumerate(G):
            for i in mix["aaSeqCDR3"]:
                if i == g:
                    node = math.sqrt(mix.loc[mix[region_string] == i, "cloneFraction"].values[0])
                    nodesize.append(node)
                    color_values.append(mix.loc[mix[region_string] == i, antigen].values[0])
                    if node * 500 > 1 / 100 * 500:
                        label_numbers[g] = index
                        label_sequences[g] = g
                    else:
                        label_numbers[g] = ""
                        label_sequences[g] = ""
                    break
        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)



        cm = plt.get_cmap(preferred_cmap)
        vmin = 0
        vmax = max(color_values)
        print(max(color_values))

        partition = community.best_partition(G)
        pos = nx.spring_layout(G)

        nodes = nx.draw_networkx_nodes(G,
                                       pos,
                                       node_size=nodesize,
                                       cmap=cm,
                                       node_color=color_values,
                                       ax=ax,
                                       vmin=vmin,
                                       vmax=vmax)
        nx.draw_networkx_edges(G, pos, alpha=0.3)
        if label_type == "numbers":
            nx.draw_networkx_labels(G, pos,
                                    labels=label_numbers,
                                    alpha=0.5,
                                    font_color="grey",
                                    horizontalalignment="right",
                                    verticalalignment="bottom", ax=ax)
        elif label_type == "sequence":
            nx.draw_networkx_labels(G, pos,
                                    labels=label_sequences,
                                    alpha=0.5,
                                    font_color="grey",
                                    horizontalalignment="right",
                                    verticalalignment="bottom", ax=ax)

        cluster_report_inter = pd.DataFrame([partition]).T
        cluster_report_inter[f'Sequences_{experiment}'] = cluster_report_inter.index
        cluster_report_inter.reset_index(drop=True, inplace=True)
        cluster_report_inter.rename(columns={0: f'Clusters_{experiment}_No.'}, inplace=True)
        cluster_report = pd.concat([cluster_report, cluster_report_inter])
        ax.set_title(experiment)
        sm = plt.cm.ScalarMappable(cmap=preferred_cmap,
                                   norm=plt.Normalize(vmin=vmin, vmax=vmax))
        plt.colorbar(sm, ax = ax)
        plot_number = plot_number + 1
  #  fig.colorbar(nodes)
    return cluster_report


### is not in use
def cluster_antigens(sequencing_report, sample, antigens, batch_size):
    report_batch = sequencing_report.groupby("Experiment").head(batch_size)
    Tot = len(antigens)
    Cols = int(input("How many Columns for the Plots do you want?"))
    # Compute Rows required
    Rows = Tot // Cols
    #     EDIT for correct number of rows:
    #     If one additional row is necessary -> add one:
    if Tot % Cols != 0:
        Rows += 1
    # Create a Position index
    Position = range(1, Tot + 1)
    plot_number = 0
    fig = plt.figure(1)
    max_colorbar = 0
    batch = report_batch[report_batch["Experiment"] == sample]
    mix = pd.concat([batch, delfia_samples])
    mix = mix.fillna(0)
    mix = mix.reset_index()
    mix = mix.drop("index", axis=1)
    G, degree_sequence = cleaning(mix, max_ld, min_ld, region_string)
    G_deg = G.degree()
    to_remove = [n for (n, deg) in G_deg if deg == 0]
    G.remove_nodes_from(to_remove)
    for antigen in antigens:
        ax = fig.add_subplot(Rows, Cols, Position[plot_number])
        nodesize = []
        clone_counts = mix.readCount
        color_values = []

        for i in range(mix.shape[0]):
            for g in G:
                if i == g:
                    node = math.sqrt(clone_counts.iloc[i])
                    nodesize.append(node)
                    color_values.append(mix[antigen].iloc[i])
        for y, n in enumerate(nodesize):
            if n > 0 and n < 15:
                nodesize[y] = 15
            if n == 0:
                nodesize[y] = 50

        colormap = plt.get_cmap("Blues")
        vmin = min(color_values)
        vmax = max(color_values)
        if vmax > max_colorbar:
            max_colorbar = vmax
        #   net = networkx.draw_networkx(G, node_size=nodesize, node_color = color_values, cmap = cm,with_labels=False, edgecolors = "Black", ax = ax)
        net = networkx.draw_networkx_nodes(G,
                                           pos=networkx.spring_layout(G),
                                           node_size=nodesize,
                                           node_color=color_values,
                                           cmap=colormap,
                                           edgecolors="Black")
        ax.set_title(antigen)
        #   sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=0, vmax=max_colorbar))
        #  cbar = plt.colorbar(sm)
        plot_number = plot_number + 1
    fig.suptitle("Clustering for sample " + sample)
#    sm = plt.cm.ScalarMappable(cmap=colormap,
#                              norm=plt.Normalize(vmin = vmin, vmax=max_colorbar))

# fig.colorbar(ax, col_ax)








import os
import subprocess
import tarfile
import requests

class NetSolP:
    
    def launch():
        save_dir =  os.path.join(os.path.dirname(__file__), "NetSolP-1.0", "model_weights")
        os.makedirs(save_dir, exists_ok = True)
        response = requests.get(r"https://services.healthtech.dtu.dk/services/NetSolP-1.0/netsolp-1.0.ALL.tar.gz")
        if response.status_code == 200:
            # Open the file and write the content in binary mode
            with open(os.path.join(), "wb") as f:
                f.write(response.content)
            print("File downloaded successfully and saved as:", filename)

            # Extract the contents of the downloaded file
            extract_dir = os.path.join(save_directory, "extracted")
            os.makedirs(extract_dir, exist_ok=True)
            with tarfile.open(filename, "r:gz") as tar:
                tar.extractall(extract_dir)

    
        


from Bio import SeqIO
import re



class Fasta:
    def make_fasta(sequence, sequence_name, fasta_name):

        fasta_record = SeqIO.SeqRecord(
            seq=sequence,
            id=sequence_name,
            description="",
        )

        with open(f"{fasta_name}.fasta", "a") as f:
            SeqIO.write(fasta_record, f, "fasta")
        
    
        
    def check_fasta(self, fasta_file):
        """Checks if a FASTA file contains amino acid sequences.

        Args:
            fasta_file: The path to the FASTA file.

        Returns:
            True if the FASTA file contains amino acid sequences, False otherwise.
        """


        fasta_records = SeqIO.parse(fasta_file, "fasta")

        # Check the sequence characters
        for fasta_record in fasta_records:
            
            sequence = fasta_record.seq
            print(sequence)
            if not all(x in ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L",
                                "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"]
                        for x in sequence):
                return False


        try:
            # Read the FASTA file
            fasta_records = SeqIO.parse(fasta_file, "fasta")

            # Check all of the headers
            for fasta_record in fasta_records:
                header = fasta_record.id
                print(header)
                if not self.is_valid_fasta_header(header):
                    return False

        except Exception as e:
                # If there is an exception, the FASTA file is not valid
            return False

        return True
    
    @staticmethod
    def is_valid_fasta_header(header):
        """Checks if a FASTA header is valid.

        Args:
            header: The FASTA header to check.

        Returns:
            True if the FASTA header is valid, False otherwise.
        """

        # Check if the header starts with a greater than sign
   #     if not header.startswith(">"):
   #         return False

        # Check if the header contains any whitespace characters
        if re.search(r"\s+", header) is not None:
            return False

        # Check if the header contains any invalid characters
        if re.search(r"[^\w\-]", header) is not None:
            return False

        return True
    
    
    




import subprocess
from ast import literal_eval
import pickle
import pandas as pd
import os
from .create_mixcr_commands import get_basename 
import sys
#psutil
try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass
from ..augment_data.create_mixcr_commands import CreateCommand
from ..augment_data.collect_fastqs import CollectFastq
from ..settings.check_dirs import check_dirs

def check_mixcr(path_to_mixcr, data, settings_dir, testing = False):
    if path_to_mixcr == "":
        while True:
            print("choose the mixcr java file with the file chooser")
            try:
                path_to_mixcr = filedialog.askopenfilename()
            except:
                path_to_mixcr = input("copy and paste the path to the mixcr jar file here.")
                path_to_mixcr = os.path.normpath(path_to_mixcr)
            if testing == True:
                break

            try:
                print("Test Mixcr")
                subprocess.run(["java",
                                "-jar",
                                path_to_mixcr,
                                "--version"
                                ])
                confirmation = True
            except:
                print("Apparently you have not choosen the right path to the mixcr .jar file. Please try again")
                confirmation = False
            if confirmation == True:
                path_to_mixcr = os.path.abspath(path_to_mixcr)
                data["mixcr_path"] = path_to_mixcr
                with open(settings_dir, "w") as f:
                    f.write(str(data))
                break
    else:
        pass
    return path_to_mixcr

def process_mixcr(experiment, method, testing, paired_end_sequencing, add_args):
    module_dir = os.path.abspath("")
    check_dirs(module_dir, experiment)
    if not testing:
        Collect = CollectFastq(paired_end_sequencing)
        Collect.get_files()
        files = Collect.paired
        if paired_end_sequencing:
            while True:
                if len(files) > 0:
                    Collect.call_pairs()
                    correct_matches = input("Are the matches of forward and backward reads correct? Y/n")
                else:
                    correct_matches = "n"
                    print("You will need to restart the collection of the files if you want to proceed with paired end sequencing data.\nMake sure that you have the forward and backward files in two seperate folder and that you have the same number of files.")
                if correct_matches in ["Y", "y", "N", "n"]:
                    if correct_matches in ["Y", "y"]:
                        break
                    else:
                        while True:
                            choose_strat = input("You can restart the collection of the files or manually add them. \nPress 1 for restarting the collection by choosing the directories.\nPress 2 for restarting the collection and manually choose the forward and backward reads.")
                            if choose_strat in ["1", "2"]:
                                Collect = CollectFastq(paired_end_sequencing)
                                if choose_strat == "1":
                                    Collect.get_files()
                                    files = Collect.paired
                                    break
                                else:
                                    Collect.manually_match_pairs()
                                    files = Collect.paired
                                    break
                            else:
                                print("Please enter a valid value")
                else:
                    print("Please enter a valid value")
        else:
            pass
    else:
        return
    settings_dir = os.path.join(module_dir,
                                "settings",
                                "global_vars.txt")
    with open(settings_dir) as f:
        data = f.read()
    data = literal_eval(data)
    path_to_mixcr = data["mixcr_path"]
    path_to_mixcr = check_mixcr(path_to_mixcr, data, settings_dir)

    sequencing_report = pd.DataFrame([])
    while True:
        java_heap_size = input(
            "Enter the amount of memory you want to allocate for java in Mb. Normally 1000 is sufficient.")
        try:
            java_heap_size = int(java_heap_size)
            break
        except:
            print("Please enter the amount as integer")
    while True:
        mixcr_chain = input("\n\nPress enter if you want to choose the default value with heavy chain (IGH) only. Please, type\nIG: For all immunoglobulin chains\n   IGL: For Immunoglobulin lambda locus\n    IGK: For Immunoglobulin kappa locus \nTCR: For all T-cell receptor chains\n     TRA: Only TCR alpha\n   TRB: Only TCR beta.\n Note: If you have not sequenced the coresponding region the pipeline will fail to create the sequencing report.")
        if mixcr_chain == "" or mixcr_chain in ["IG", "IGH", "IGL", "IGK", "TCR", "TRA", "TRB", ]:
            if mixcr_chain == "":
                mixcr_chain = "IGH"
            break
        else:
            print("Please enter a correct value.")
    for filename in files:
        basename = get_basename(filename) # check if basename is correct and aligns to standards, e.g. 22 char
        if " " in basename:
            print("Spaces in the filename are not allowed. Analysis aborted")
            sys.exit(1)
    
    for filename in files:
        for file in os.listdir(os.path.join(module_dir, "temp")):
            os.remove(os.path.join(module_dir,
                                   "temp",
                                   file))
        Commands = CreateCommand(module_dir, path_to_mixcr, paired_end_sequencing, experiment, filename, java_heap_size)
        subprocess.run(Commands.prepare_align(method, add_args))
        subprocess.run(Commands.prepare_assembly())
        subprocess.run(Commands.prepare_clones(mixcr_chain = mixcr_chain))
        basename = Commands.basename
        try:
            columns_not_wanted = ['refPoints', 'allVHitsWithScore', 'allDHitsWithScore',
                                  'allJHitsWithScore', 'allCHitsWithScore', 'allVAlignments',
                                  'allDAlignments', 'allJAlignments', 'allCAlignments']
            clones_sample = pd.read_table(Commands.table_tsv)
            clones_sample.drop(columns=columns_not_wanted, inplace=True)
            clones_sample["Experiment"] = basename
            sequencing_report = pd.concat([sequencing_report, clones_sample])

            for file in os.listdir(os.path.join(module_dir, "temp")):
                os.remove(os.path.join(module_dir,
                                       "temp",
                                       file))
        except:
            print(f"Processing for {basename} failed")

    if not testing:
        
        if sequencing_report.shape[0] == 0:
            raise Exception(
                "The processing failed for all of your files. Most likely because you chose the wrong mixcr method.")
        report_dir = os.path.join(module_dir,
                                  "my_experiments",
                                  experiment,
                                  "sequencing_report.csv")
        sequencing_report.to_csv(report_dir)
        data["last_experiment"] = experiment
        glob_vars_dir = os.path.join(module_dir,
                                     "settings",
                                     "global_vars.txt")
        with open(glob_vars_dir, "w") as f:
            f.write(str(data))
        unique_experiments = sequencing_report["Experiment"].unique()
        experiment_dic = {item: item for item in list(unique_experiments)}
        exp_names_dir = os.path.join(module_dir,
                                     "my_experiments",
                                     experiment,
                                     "experiment_names.pickle")
        with open(exp_names_dir, "wb") as f:
            pickle.dump(experiment_dic, f)
    print(
        "The output of your processing is a sequencing_report which contains the sequences and their sample names of all fastq files. Further, a folder with the alignment reports for each fastq file was created. You can have a look under ~/my_experiments/YOUR_EXPERIMENT_NAME")

class RunMixcr():
    def __init__(self, experiment, method, paired_end_sequencing, add_args):
        self.module_dir = os.path.abspath("")
        self.experiment = experiment
        check_dirs(self.module_dir, self.experiment)
        path_to_mixcr, data = self.validate_mixcr(self.module_dir)
        self.files = self.prep_files(paired_end_sequencing)
        sequencing_report = self.process(path_to_mixcr, method, paired_end_sequencing, add_args)
        self.tidy(sequencing_report, data)
        
        
    @staticmethod
    def prep_files(paired_end_sequencing):
        Collect = CollectFastq(paired_end_sequencing)
        Collect.get_files()
        files = Collect.paired
        if paired_end_sequencing:
            while True:
                if len(files) > 0:
                    Collect.call_pairs()
                    correct_matches = input("Are the matches of forward and backward reads correct? Y/n")
                else:
                    correct_matches = "n"
                    print("You will need to restart the collection of the files if you want to proceed with paired end sequencing data.\nMake sure that you have the forward and backward files in two seperate folder and that you have the same number of files.")
                if correct_matches in ["Y", "y", "N", "n"]:
                    if correct_matches in ["Y", "y"]:
                        break
                    else:
                        while True:
                            choose_strat = input("You can restart the collection of the files or manually add them. \nPress 1 for restarting the collection by choosing the directories.\nPress 2 for restarting the collection and manually choose the forward and backward reads.")
                            if choose_strat in ["1", "2"]:
                                Collect = CollectFastq(paired_end_sequencing)
                                if choose_strat == "1":
                                    Collect.get_files()
                                    files = Collect.paired
                                    break
                                else:
                                    Collect.manually_match_pairs()
                                    files = Collect.paired
                                    break
                            else:
                                print("Please enter a valid value")
                else:
                    print("Please enter a valid value")
        else:
            pass
        return files

    @staticmethod
    def validate_mixcr(module_dir):
        settings_dir = os.path.join(module_dir,
                                "settings",
                                "global_vars.txt")
        with open(settings_dir) as f:
            data = f.read()
        data = literal_eval(data)
        path_to_mixcr = data["mixcr_path"]
        path_to_mixcr = check_mixcr(path_to_mixcr, data, settings_dir)
        return path_to_mixcr, data
    
    @staticmethod
    def ask_heap_size():
        while True:
            java_heap_size = input(
                "Enter the amount of memory you want to allocate for java in Mb. Normally 1000 is sufficient.")
            try:
                java_heap_size = int(java_heap_size)
                break
            except:
                print("Please enter the amount as integer")
        return java_heap_size
    
    @staticmethod
    def ask_chain():
        while True:
            mixcr_chain = input("\n\nPress enter if you want to choose the default value with heavy chain (IGH) only. Please, type\nIG: For all immunoglobulin chains\n   IGL: For Immunoglobulin lambda locus\n    IGK: For Immunoglobulin kappa locus \nTCR: For all T-cell receptor chains\n     TRA: Only TCR alpha\n   TRB: Only TCR beta.\n Note: If you have not sequenced the coresponding region the pipeline will fail to create the sequencing report.")
            if mixcr_chain == "" or mixcr_chain in ["IG", "IGH", "IGL", "IGK", "TCR", "TRA", "TRB", ]:
                if mixcr_chain == "":
                    mixcr_chain = "IGH"
                break
            else:
                print("Please enter a correct value.")
        return mixcr_chain
    
    def check_basenames(self):    
        for filename in self.files:
            basename = get_basename(filename) # check if basename is correct and aligns to standards, e.g. 22 char
            if " " in basename:
                print("Spaces in the filename are not allowed. Analysis aborted")
                sys.exit(1)
    
    
    def process(self,path_to_mixcr,  method, paired_end_sequencing, add_args):         
        
        java_heap_size = self.ask_heap_size()
        mixcr_chain = self.ask_chain()
        
        sequencing_report = pd.DataFrame([])        
        for filename in self.files:
            for file in os.listdir(os.path.join(self.module_dir, "temp")):
                os.remove(os.path.join(self.module_dir,
                                    "temp",
                                    file))
            Commands = CreateCommand(self.module_dir, path_to_mixcr, paired_end_sequencing, self.experiment, filename, java_heap_size)
            subprocess.run(Commands.prepare_align(method, add_args))
            subprocess.run(Commands.prepare_assembly())
            subprocess.run(Commands.prepare_clones(mixcr_chain = mixcr_chain))
            basename = Commands.basename
            try:
                columns_not_wanted = ['refPoints', 'allVHitsWithScore', 'allDHitsWithScore',
                                    'allJHitsWithScore', 'allCHitsWithScore', 'allVAlignments',
                                    'allDAlignments', 'allJAlignments', 'allCAlignments']
                clones_sample = pd.read_table(Commands.table_tsv)
                clones_sample.drop(columns=columns_not_wanted, inplace=True)
                clones_sample["Experiment"] = basename
                sequencing_report = pd.concat([sequencing_report, clones_sample])

                for file in os.listdir(os.path.join(self.module_dir, "temp")):
                    os.remove(os.path.join(self.module_dir,
                                        "temp",
                                        file))
            except:
                print(f"Processing for {basename} failed")
                
        return sequencing_report
    
    def tidy(self, sequencing_report, data):
        if sequencing_report.shape[0] == 0:
            raise Exception(
                "The processing failed for all of your files. Most likely because you chose the wrong mixcr method.")
        report_dir = os.path.join(self.module_dir,
                                  "my_experiments",
                                  self.experiment,
                                  "sequencing_report.csv")
        sequencing_report.to_csv(report_dir)
        data["last_experiment"] = self.experiment
        glob_vars_dir = os.path.join(self.module_dir,
                                     "settings",
                                     "global_vars.txt")
        with open(glob_vars_dir, "w") as f:
            f.write(str(data))
        unique_experiments = sequencing_report["Experiment"].unique()
        experiment_dic = {item: item for item in list(unique_experiments)}
        exp_names_dir = os.path.join(self.module_dir,
                                     "my_experiments",
                                     self.experiment,
                                     "experiment_names.pickle")
        with open(exp_names_dir, "wb") as f:
            pickle.dump(experiment_dic, f)
        print(
            "The output of your processing is a sequencing_report which contains the sequences and their sample names of all fastq files. Further, a folder with the alignment reports for each fastq file was created. You can have a look under ~/my_experiments/YOUR_EXPERIMENT_NAME")



import subprocess
import os

class VE:
    
    def __init__(self, name_ve):
        self.name_ve
        self.create_ve()
        self.switch_to_ve()
        
    
    def check_path(self):
        wd = os.getcwd()
        ve_existence = os.path.isdir(os.path.join(wd, self.name_ve))
        return ve_existence
    
    def check_venv_correct(self):
        wd = os.getcwd()
        exists = os.path.isfile(os.path.join(wd, self.name_ve, "pyvenv.cfg"))
        return exists
    
    def create_ve(self):
        if self.check_path() == False:
            subprocess.run(["python",
                        "-m",
                        "venv",
                        self.name_ve])
        elif self.check_venv_correct() == False:
            subprocess.run(["python",
                            "-m",
                            "venv",
                            self.name_ve + "_ve"])        
    def switch_to_ve(self):
        subprocess.run([os.path.join(self.name_ve, "Scripts", "activate")])
        
    def leave_ve(self):
        subprocess.run(["deactivate"])


import networkx as nx
import math
from textwrap import wrap
import numpy as np
from ..tidy_protbert_embedding import TransformerBased
import matplotlib.pyplot as plt
import community

class Network_Embedding:
    def __init__(self, ax, sequencing_report, list_experiments, batch_size,font_settings,cmap = plt.cm.RdYlBu, nodesize = None, threshold_distance = 1) -> None:
        Transformer = TransformerBased(choice = "protbert")
        sequences,sequences_filtered,  selected_rows = Transformer.filter_sequences(sequencing_report, batch_size=batch_size, experiments = list_experiments, region_of_interest="aaSeqCDR3", )
        sequences_filtered = sequences_filtered.to_list()
        sequences_list = Transformer.embedding_per_seq(sequences, normalize = True)
        
        self.G = nx.Graph()
        self.add_edges(sequences_list, sequences_filtered, threshold_distance)
        if nodesize == None:
            nodesize = self.add_nodesize(self.G, selected_rows)
        else:
            nodesize = 30
        self.generate_plot(self.G, nodesize, ax, cmap)
        title = "\n".join(wrap("t-SNE embedding of for given samples", 40))
        self.ax.set_title(title, pad= 12, **font_settings)

    def add_edges(self, sequences_list, sequences_filtered, threshold_distance):
        num_sequences = sequences_list.shape[0]
        for seq_1 in range(num_sequences):
            for seq_2 in range(seq_1 + 1, num_sequences):
                p = sequences_list[seq_1]
                q = sequences_list[seq_2]
                distance = np.sum(abs(sequences_list[seq_1] - sequences_list[seq_2]))
                print(distance)
                if distance < threshold_distance/num_sequences:
                    string1 = sequences_filtered[seq_1]
                    string2 = sequences_filtered[seq_2]
                    self.G.add_edge(string1, string2, weight=distance)
    @staticmethod
    def add_nodesize(G, selected_rows):
        nodesize = []
        label_numbers = {}
        label_sequences = {}
        for index, g in enumerate(G):
            
            for i in selected_rows["aaSeqCDR3"]:
                
                if i == g:
                    node = math.sqrt(selected_rows.loc[selected_rows["aaSeqCDR3"] == i, "readFraction"].values[0])
                    nodesize.append(node)
                    if node * 500 > 1 / 100 * 500:
                        label_numbers[g] = index
                        label_sequences[g] = g
                    else:
                        label_numbers[g] = ""
                        label_sequences[g] = ""
                    break
        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)
        return nodesize
    
    @staticmethod     
    def generate_plot(G, nodesize, ax, cmap):
        partition = community.best_partition(G)
        pos = nx.spring_layout(G)
        nx.draw_networkx_nodes(G, pos,node_size=nodesize, cmap=cmap, node_color=list(partition.values()), ax=ax)
        nx.draw_networkx_edges(G, pos, alpha=0.3, ax=ax)







import squarify
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
from textwrap import wrap


class VisFrac:

    def __init__(
        self,
        sequencing_report,
        ax,
        sample,
        visualize_sequences,
        prefered_cmap,
        top_clone_fraction,
        seqs_viz_fraction,
        alpha_val,
        pad_rectangles,
        force_reducing,
        limit_seq_filter,
        font_settings,
        region_string,
    ):
        self.ax = ax
        top_95_percent = self.get_top_fraction(
            report=sequencing_report,
            sample=sample,
            force_reducing=force_reducing,
            fraction=top_clone_fraction,
        )
        self.seqs = self.prepare_labels(
            top_95_percent, region_string, limit_seq_filter, seqs_viz_fraction
        )
        self.fracs = top_95_percent["cloneFraction"]
        self.make_plot(
            pad=pad_rectangles,
            visualize_seqs=visualize_sequences,
            alpha_val=alpha_val,
            prefered_cmap=prefered_cmap,
        )
        self.no_axis()
        self.add_title(font_settings, sample, top_clone_fraction)

    @staticmethod
    def get_top_fraction(report, sample, force_reducing, fraction=0.95):
        report = report[report["Experiment"] == sample]
        index = report["cloneFraction"].cumsum() < fraction
        top_95_percent = report[index]
        if top_95_percent.shape[0] > 1000:
            if force_reducing != None:
                top_95_percent = top_95_percent.iloc[:force_reducing]
        top_95_percent = top_95_percent.copy()
        top_95_percent["cloneFraction"] = (
            top_95_percent["cloneFraction"] / fraction
        )  # normalize
        return top_95_percent

    @staticmethod
    def prepare_labels(report, region_string, limit_seq_filter, fraction=0.75):
        index = report["cloneFraction"].cumsum() < fraction
        seqs = report[region_string]
        seq_length = len(seqs)
        filtered_seqs = seqs[index]
        filtered_length = len(filtered_seqs)
        if filtered_length > limit_seq_filter:
            filtered_seqs = filtered_seqs.iloc[:limit_seq_filter]
            filtered_length = len(filtered_seqs)
        empty_values = pd.Series(
            [""] * int((seq_length - filtered_length)), dtype="str"
        )
        new_seqs = pd.concat([filtered_seqs, empty_values], ignore_index=True)
        return new_seqs

    @staticmethod
    def prep_cmap(top_readF, prefered_cmap="Reds"):
        cmap = matplotlib.colormaps.get_cmap(prefered_cmap)
        mini = min(top_readF)
        maxi = max(top_readF)
        norm = matplotlib.colors.Normalize(vmin=mini, vmax=maxi)
        colors = [cmap(norm(value)) for value in top_readF]
        return colors

    def no_axis(self):
        self.ax.set_axis_off()

    def add_title(self, font_settings, sample, top_clone_fraction):
        title = "\n".join(
            wrap("Clone fraction of " + sample + f"for top {top_clone_fraction} %", 40)
        )
        self.ax.set_title(title, pad=12, **font_settings)

    def make_plot(self, pad, visualize_seqs, alpha_val, prefered_cmap, **kwargs):
        if visualize_seqs == False:
            self.seqs = None
        else:
            pass
        colors = self.prep_cmap(self.fracs, prefered_cmap)
        squarify.plot(
            self.fracs,
            pad=pad,
            label=self.seqs,
            alpha=alpha_val,
            color=colors,
            ax=self.ax,
            text_kwargs={"fontsize": 5},
            **kwargs,
        )



from .layout_finder import best_layout
import matplotlib.pyplot as plt

class Subplotter:
    def __init__(self):
        self.current_figure = 1
        self.all_figures = []
        self.all_axes = []
        setattr(self, f"fig{self.current_figure}", None)
        setattr(self, f"axes_all{self.current_figure}", None)
        
    

        
    def update_fig(self, figure, axes_list):
        rows, cols = best_layout(len(axes_list))
        fig, axes = plt.subplots(nrows=rows, ncols=cols)
        for ax in fig.axes:
            fig.delaxes(ax)

        # Add new axes from axes_list
        for ax in axes_list:
            fig._axstack.add(fig._make_key(ax), ax)
        # modify class attributes
        if hasattr(self, f"fig{figure}"):
            given_fig = getattr(self, f"fig{figure}")
        else:
            setattr(self, f"fig{figure}", fig)
            setattr(self, f"axes_all{figure}", axes)
            given_fig = getattr(self, f"fig{figure}")
            
        if given_fig in self.all_figures:
            pass
        else:
            self.all_figures.append(getattr(self, f"fig{figure}"))
   #     self.all_axes.append(getattr(self, f"axes_all{figure}"))

        
    
    
    def set_attribute(self, figure):
        if not hasattr(self, f"axes{figure}"):
            setattr(self, f"axes{figure}", [])
        else:
            pass
        
    def add_to_subplots(self, ax, figure = 1,):
        if self.current_figure == figure:
            pass
        else:
            self.current_figure = figure
        
        self.set_attribute(figure)
        axes_list = getattr(self, f"axes{figure}")
        axes_list.append(ax)
        self.update_fig(figure, axes_list)
        
     
     
SubPlots = Subplotter()
     
import numpy as np

# Data for the plots
x = np.linspace(0, 10, 100)
y1 = np.sin(x)
y2 = np.cos(x)
y3 = x ** 2   

        
fig1, ax1 = plt.subplots()
ax1.plot(x, y1)
ax1.set_title('Line Plot')
ax1.set_xlabel('x')
ax1.set_ylabel('y')

SubPlots.add_to_subplots(ax1)

# Plot 2: Scatter plot
fig2, ax2 = plt.subplots()
ax2.scatter(x, y2)
ax2.set_title('Scatter Plot')
ax2.set_xlabel('x')
ax2.set_ylabel('y')

SubPlots.add_to_subplots(ax2)

# Plot 3: Bar plot
fig3, ax3 = plt.subplots()
ax3.bar(x, y3)
ax3.set_title('Bar Plot')
ax3.set_xlabel('x')
ax3.set_ylabel('y')

SubPlots.add_to_subplots(ax3)
SubPlots.create_figure()
plt.show()



from textwrap import wrap
from .tidy_protbert_embedding import TransformerBased
import pandas as pd
import matplotlib.pyplot as plt
import warnings
import numpy as np
import os
from .contents.simple_protein_property import GetProteinProperty


class PrepareData:
    def __init__(self):

        self.tsne_results = None
        self.clones = None

    @staticmethod
    def logical_check(batch_size, perplexity, pca_components, model, iterations_tsne):
        # in the pipeline is another test, where the sample name is checked which has to be in the sequencing report
        assert (
            batch_size > perplexity
        ), "The batch_size has to be larger than the perplexity"
        assert (
            batch_size > pca_components
        ), "The batch_size has to be larger than the pca_components"
        assert (
            batch_size != 0
        ), "batch_size value must not be 0"  # indirectly tests that perplexity and pca components must not be 0 as well
        # more models will follow
        models_all = [
            "Rostlab/ProstT5_fp16",
            "Rostlab/prot_t5_xl_uniref50",
            "Rostlab/prot_t5_base_mt_uniref50",
            "Rostlab/prot_bert_bfd_membrane",
            "Rostlab/prot_t5_xxl_uniref50",
            "Rostlab/ProstT5",
            "Rostlab/prot_t5_xl_half_uniref50-enc",
            "Rostlab/prot_bert_bfd_ss3",
            "Rostlab/prot_bert_bfd_localization",
            "Rostlab/prot_electra_generator_bfd",
            "Rostlab/prot_t5_xl_bfd",
            "Rostlab/prot_bert",
            "Rostlab/prot_xlnet",
            "Rostlab/prot_bert_bfd",
            "Rostlab/prot_t5_xxl_bfd",
        ]
        assert (
            model in models_all
        ), f"Please enter a valid model name which are\n{models_all}. You can find the models at: https://huggingface.co/Rostlab"
        assert (
            iterations_tsne > 250
        ), "The number of iterations must be larger than 250 according to sklearn"
        pca_components > 2, "You must insert more than two pca components"

    @staticmethod
    def datatype_check(
        samples, pca_components, perplexity, iterations_tsne, batch_size, model
    ):
        assert (
            type(samples) == list
        ), "You have to give a list with the samples you want to analyze"
        assert (
            type(pca_components) == int
        ), "You have to give an integer as input for the pca_components"
        assert (
            type(perplexity) == int
        ), "You have to give an integer as input for the perplexity"
        assert (
            type(iterations_tsne) == int
        ), "You have to give an integer as input for the iterations_tsne"
        assert (
            type(batch_size) == int
        ), "You have to give an integer as input for the batch_size"
        assert type(model) == str, "The input for model must be a string"

    @staticmethod
    def check_warnings(sequences, pca_components, perplexity):
        if len(sequences) < pca_components:
            warnings.warn(
                f"The number of sequences you have is {len(sequences)} but you need to have more sequences than principal components which is: {pca_components}"
            )
            print(
                f"Number of principal components is set to number of sequences ({len(sequences)})"
            )
            pca_components = len(sequences)
        if pca_components < perplexity:
            warnings.warn(
                "The number of reduced dimensions you have is "
                + str(pca_components)
                + "but you need to have more than perplexity which is: "
                + str(perplexity)
            )
            print(
                f"Perplexity is set to the half of reduced dimensions ({pca_components//2})"
            )
            perplexity = pca_components // 2
            if perplexity < 1:
                perplexity = 1
        return pca_components, perplexity

    @staticmethod
    def return_binding_results(
        selected_rows, antigens, region_of_interest, add_clone_size, tsne_results
    ):
        if antigens is not None:
            kds = selected_rows[antigens].max(
                axis=1
            )  # if there are multiple values for the same sequence this will find the highest one
            tsne_results["binding"] = list(kds)
            ids = selected_rows[antigens].idxmax(axis=1)
            tsne_results["highest_binder"] = list(ids)
        else:
            kds = None
            ids = None
        aminoacids = selected_rows[region_of_interest].to_list()
        experiments_batch = selected_rows["Experiment"]
        experiments_batch = experiments_batch.replace(0, "non-merged binding data")
        unique_experiments_num = pd.factorize(experiments_batch)[0]
        tsne_results["experiments_string"] = experiments_batch.to_list()
        tsne_results["experiments_factorized"] = unique_experiments_num
        tsne_results["sequences"] = list(aminoacids)
        tsne_results["sequence_id"] = list(range(tsne_results.shape[0]))
        if add_clone_size != None:
            tsne_results["size"] = (
                np.array(selected_rows["cloneFraction"].to_list()) * add_clone_size
            )
        else:
            tsne_results["size"] = (
                len(selected_rows["cloneFraction"].to_list()) * [300]
            )
        return kds, ids, tsne_results

    @staticmethod
    def filter_binding_data(binding_data, region_of_interest, antigens):
        if binding_data is not None:
            merged_columns = [region_of_interest] + antigens
            binding_data = binding_data[merged_columns]
        return binding_data

    def label_sequence_characteristic(
        self, characteristic, sequences, selected_rows, antigens
    ):
        if characteristic != None:
            if characteristic != "binding":
                Property = GetProteinProperty(sequences)
                Property.calc_attribute(attribute=characteristic)
                self.tsne_results = Property.add_attributes_to_table(
                    self.tsne_results, attribute=characteristic
                )
                property_result = list(Property.sequence_property_interest.values())
            else:
                kds = selected_rows[antigens].max(axis=1)
                property_result = kds.fillna(0)
        else:
            property_result = None
            pass
        return property_result

    @staticmethod
    def save_current_embedding(
        X_path: str,
        sequences_list: np.array,
        batch_size: int,
        antigens,
        region_of_interest: str,
        model_choice: str,
        binding_data: pd.DataFrame,
        iterations: int
    ):
        """This method saves the sequences_list from which contains the array with the high dimensions as npz file.
        Further, it creates in the same directory a pkl file which contains the parameters for that specific embedding.


        Args:
            X_path (str): Path where yhe array should be saved
            sequences_list (np.array): array with the data from the embedding.
            batch_size (int): Number of sequences per sample
            antigens (list): Value for antigens which should be incorporated. either none or list of strings
            region_of_interest (str): Region of interest
            model_choice (str): Model which was chosen for the embedding
            binding_data (pd.DataFrame): pandas dataframe with the embedding
        """
        np.savez_compressed(X_path, a=sequences_list)
        current_run = {
            "batch_size": batch_size,
            "antigens": antigens,
            "iterations": iterations,
            "region_of_interest": region_of_interest,
            "model_choice": model_choice,
            "binding_data": binding_data,
        }
        dir_X = os.path.dirname(X_path)
        np.save(os.path.join(dir_X, "current_run.npy"), current_run)

    @staticmethod
    def load_and_compare_past_run(
        X_path,
        batch_size: int,
        antigens,
        region_of_interest: str,
        model_choice: str,
        binding_data: pd.DataFrame,
        iterations:int
    ):
        """Loads the np array where the embedding is saved and compares the settings of the past run to the settings of the current run and if any is different res will be False and the embedding will be repeated.

        Args:
            X_path (_type_): _description_
            batch_size (int): _description_
            antigens (_type_): _description_
            region_of_interest (str): _description_
            model_choice (str): _description_
            binding_data (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        saved_array = np.load(X_path)
        sequences_list = saved_array["a"]
        dir_X = os.path.dirname(X_path)
        try:
            old_settings = np.load(
                os.path.join(dir_X, "current_run.npy"), allow_pickle=True
            ).item()

        except:
            res = False
            return res, sequences_list
        new_settings = {
            "batch_size": batch_size,
            "antigens": antigens,
            "iterations": iterations,
            "region_of_interest": region_of_interest,
            "model_choice": model_choice,
            "binding_data": binding_data,
        }
        binding_data_old = old_settings.get("binding_data")
        binding_data_new = new_settings.get("binding_data")
        new_settings.pop("binding_data")
        old_settings.pop("binding_data")
        if binding_data is None:
            res = False
            return res, sequences_list
        res = all(
            (new_settings.get(k) == v for k, v in old_settings.items())
        ) and binding_data_old.equals(binding_data_new)
        return res, sequences_list

    def tidy(
        self,
        sequencing_report,
        list_experiments,
        region_of_interest,
        antigens=None,
        batch_size=300,
        pca_components=70,
        perplexity=25,
        iterations_tsne=1000,
        eps_dbscan=0.5,
        min_pts_dbscan=2,
        add_clone_size=500,
        model_choice="Rostlab/prot_bert",
        characteristic=None,
        binding_data=None,
        cf_column_name="cloneFraction",
        sample_column_name="Experiment",
        X_path="temp/current_array.npz",
    ):
        """creates tsne_results as class object which is a table containing all necessary data for the final plot

        Args:
            sequencing_report (pd.DataFrame): report which contains all sequencing information from all samples after the processing with mixcr
            list_experiments (list): List containing the names of the samples which should be used for the final plot.
            region_of_interest (str): A string which indicates the column name from which the amino acid sequences should be taken from.
            antigens (list, optional): list containing the names of the antigens which are the headers of the binding data. From these antigens the binding data will be taken for the plot. Defaults to None.
            batch_size (int, optional): Equals to the number of sequences which are drawn from the sequencing report for the embedding and the final plot. Defaults to 300.
            pca_components (int, optional): Equals to the number of principal components which are used as input for tsne. This helps to remove noise and complexity before using t-SNe. Defaults to 70.
            perplexity (int, optional): Number for the parameter perplexity in t-SNE. Defaults to 25.
            iterations_tsne (int, optional): Equals to the number of iterations the t-SNE algorithm has to run. Defaults to 1000.
            model_choice (str, optional): Is the final model you choose to embed your sequences. Defaults to "Rostlab/prot_bert".
            binding_data (pd.DataFrame, optional): Dataframe which contains the sequences and the binding values to the antigens. Defaults to None.
            cf_column_name (str, optional): Name of the column which contains the clone fraction in the sequencing report. Defaults to "cloneFraction".
            sample_column_name (str, optional): Name of the column which contains the sample names in the sequencing report. Defaults to "Experiment".

        Returns:
            _type_: _description_
        """
        for sample in list_experiments:
            assert sample in list(
                sequencing_report[sample_column_name].unique()
            ), f"{sample} does not exist"
        if binding_data is not None:
            assert (
                region_of_interest in binding_data.columns.to_list()
            ), f"You must have sequences for {region_of_interest} in your binding data"
        batch_size = batch_size * len(list_experiments)

        self.logical_check(
            batch_size, perplexity, pca_components, model_choice, iterations_tsne
        )
        self.datatype_check(
            list_experiments,
            pca_components,
            perplexity,
            iterations_tsne,
            batch_size,
            model_choice,
        )
        binding_data = self.filter_binding_data(
            binding_data, region_of_interest, antigens
        )

        sequences, sequences_filtered, selected_rows = (
            TransformerBased.filter_sequences(
                sequencing_report,
                batch_size / len(list_experiments),
                list_experiments,
                binding_data,
                region_of_interest=region_of_interest,
                cf_column_name=cf_column_name,
                sample_column_name=sample_column_name,
            )
        )
        peptides = selected_rows[region_of_interest].to_list()
        self.clones = selected_rows[cf_column_name].to_list()

        #    pca_components, perplexity = self.check_warnings(sequences, pca_components, perplexity)

        if os.path.exists(X_path):
            res, sequences_list = self.load_and_compare_past_run(
                X_path,
                batch_size,
                antigens,
                region_of_interest,
                model_choice,
                binding_data,
                iterations_tsne
            )
        else:
            res = False

        if res == False:
            Transformer = TransformerBased(choice=model_choice)
            sequences_list = Transformer.embedding_per_seq(sequences)
            self.save_current_embedding(
                X_path,
                sequences_list,
                batch_size,
                antigens,
                region_of_interest,
                model_choice,
                binding_data,
                iterations_tsne
            )
        X = TransformerBased.do_pca(sequences_list, pca_components)
        tsne_results = TransformerBased.do_tsne(X, perplexity, iterations_tsne)
  #      get_clusters = TransformerBased.cluster_with_hdbscan( # not recommended for tsne
  #          tsne_results, eps=eps_dbscan, min_pts=min_pts_dbscan
   #     ).tolist()    
        
        property_result = self.label_sequence_characteristic(
            characteristic, sequences, selected_rows, antigens
        )
        if property_result != None:
            tsne_results[characteristic] = property_result

    #    tsne_results["cluster_id"] = get_clusters
        tsne_results["cloneFraction"] = self.clones
        kds, ids, tsne_results = self.return_binding_results(
            selected_rows, antigens, region_of_interest, add_clone_size, tsne_results
        )  # add columns to report
        for sample in list_experiments:  # tests
            assert (
                tsne_results[tsne_results["experiments_string"] == sample].shape[0] >= 1
            ), f"After processing your data for your parameters no sequences are left for {sample}"
        assert type(tsne_results["experiments_string"].tolist()) == list
        self.tsne_results = tsne_results
        return peptides, selected_rows, kds, ids

    def make_csv(self, path=None):
        if path == None:
            save_path = "tsne_results.csv"
        else:
            save_path = path
        assert path.endswith(".csv")
        self.tsne_results.to_csv(save_path)


class PlotEmbedding:
    def __init__(
        self,
        sequencing_report,
        model_choice,
        list_experiments,
        region_of_interest,
        strands = True,
        add_clone_size = 500,
        batch_size = 300,
        pca_components = 50,
        perplexity = 25,
        iterations_tsne = 1000,
        eps_dbscan=0.5,
        min_pts_dbscan=2,
        characteristic = None,
        antigens=None,
        font_settings={},
        legend_settings={},
        ax=None,
        binding_data=None,  # antigens mutual attribute
        colorbar_settings=None,
        toxin_names=None,
        extra_figure=False,
        prefered_cmap="inferno",
    ):
        self.ax = ax
        self.binding_data = binding_data
        self.data_prep = PrepareData()
        peptides, selected_rows, kds, ids = self.data_prep.tidy(
            sequencing_report=sequencing_report,
            list_experiments=list_experiments,
            region_of_interest=region_of_interest,
            antigens=antigens,
            batch_size=batch_size,
            pca_components=pca_components,
            eps_dbscan=eps_dbscan,
            min_pts_dbscan=min_pts_dbscan,
            characteristic=characteristic,
            perplexity=perplexity,
            iterations_tsne=iterations_tsne,
            model_choice=model_choice,
            binding_data=binding_data,
            add_clone_size=add_clone_size,
        )
        self.tsne_results = self.data_prep.tsne_results
        if self.ax != None:
            if self.binding_data is not None:
                self.create_binding_plot(
                    kds, ids, toxin_names, colorbar_settings, prefered_cmap
                )
                if extra_figure == True and font_settings != {}:
                    self.create_second_bind_plot(font_settings)
                    title = "\n".join(wrap(f"t-SNE embedding for {antigens}", 40))
                    self.ax.set_title(title, pad=12, **font_settings)
            else:
                title = "\n".join(wrap("TSNE embedding for given samples", 40))
                sm = self.create_plot(characteristic, prefered_cmap)
                if colorbar_settings != {} and sm != None:
                    self.add_colorbar(colorbar_settings, characteristic, sm)
                if font_settings != {}:
                    self.ax.set_title(title, pad=12, **font_settings)

            if font_settings != {}:
                self.ax.set_xlabel("t-SNE1", **font_settings)  # add font_settings
                self.ax.set_ylabel("t-SNE2", **font_settings)

            if strands == True:
                self.add_seq_anotation(peptides)
            if legend_settings != {}:
                self.add_legend(legend_settings)

    def create_plot(self, characteristic, prefered_cmap):
        markers = ["o", "+", "x", "s", "p", "x", "D"]
        if characteristic == None:
            self.tsne_results["color"] = self.tsne_results["experiments_factorized"]
            sm = None
        else:
            self.tsne_results["color"] = self.tsne_results[characteristic]
            global_min_color = self.tsne_results["color"].min()
            global_max_color = self.tsne_results["color"].max()
            sm = plt.cm.ScalarMappable(
                cmap=prefered_cmap,
                norm=plt.Normalize(vmin=global_min_color, vmax=global_max_color),
            )

            norm = plt.Normalize(vmin=global_min_color, vmax=global_max_color)

        unique_experiments = self.tsne_results["experiments_string"].unique()

        for index, experiment in enumerate(unique_experiments):
            local_results = self.tsne_results[
                self.tsne_results["experiments_string"] == experiment
            ]
            tsne_1_values = local_results["tsne1"]
            tsne_2_values = local_results["tsne2"]
            if characteristic is not None:
                self.ax.scatter(
                    tsne_1_values,
                    tsne_2_values,
                    marker=markers[index],
                    c=local_results["color"],
                    s=local_results["size"],
                    alpha=0.5,
                    norm=norm,
                    cmap=prefered_cmap,
                    label=experiment,
                )

            else:
                
                self.ax.scatter(
                    tsne_1_values,
                    tsne_2_values,
                    marker=markers[index],
                    c=local_results["color"],
                    s=local_results["size"],
                    alpha=0.5,
                    cmap="tab20c",
                    label=experiment,
                )

        return sm

    def add_colorbar(self, colorbar_settings, label, sm):
        colorbar_settings["orientation"] = "horizontal"
        del colorbar_settings["spacing"]
        fig = self.ax.get_figure()
        fig.colorbar(
            sm,
            ax=self.ax,
            label=label,
            **colorbar_settings,
        )

    def add_size(self, add_clone_size):
        size_points = self.data_prep.clones * add_clone_size
        self.tsne_plot.set_sizes(size_points)

    def add_legend(self, legend_settings):
        self.ax.legend(
            **legend_settings,
        )

    def add_seq_anotation(self, peptides):
        x = self.tsne_results["tsne1"].values.tolist()
        y = self.tsne_results["tsne2"].values.tolist()
        for i in range(0, len(x), 10):
            self.ax.annotate(
                peptides[i],
                (x[i], y[i]),
                fontsize=5,
            )

    def create_second_bind_plot(self, font_settings):
        self.fig2 = plt.figure(100)
        self.ax2 = self.fig2.gca()
        self.ax2.scatter(self.tsne_results.tsne1, self.tsne_results.tsne2, alpha=0.0)
        self.ax2.set_xlabel("t-SNE 1", **font_settings)
        self.ax2.set_ylabel("t-SNE 2", **font_settings)
        n = 0
        for j, row in self.tsne_results.iterrows():
            if row["binding"] > 1:
                self.ax2.text(
                    row["tsne1"],
                    row["tsne2"],
                    row["sequence_id"],
                    fontsize=10,
                    weight="bold",
                )

            else:
                if n == 6:
                    n = 0
                    self.ax2.text(
                        row["tsne1"], row["tsne2"], row["sequence_id"], fontsize=8
                    )
            n += 1

    def get_sm(self, column_char, prefered_cmap):
        self.tsne_results["color"] = column_char
        global_min_color = self.tsne_results["color"].min()
        global_max_color = self.tsne_results["color"].max()
        sm = plt.cm.ScalarMappable(
            cmap=prefered_cmap,
            norm=plt.Normalize(vmin=global_min_color, vmax=global_max_color),
        )

        norm = plt.Normalize(vmin=global_min_color, vmax=global_max_color)
        return sm, norm

    def create_binding_plot(
        self, kds, ids, toxin_names, colorbar_settings, prefered_cmap="magma"
    ):
        markers = ["o", "+", "x", "s", "p", "x", "D"]
        self.tsne_results["color"] = self.tsne_results["binding"]
        sm, norm = self.get_sm(self.tsne_results["binding"], prefered_cmap)

        unique_experiments = self.tsne_results["experiments_string"].unique()

        for index, experiment in enumerate(unique_experiments):
            local_results = self.tsne_results[
                self.tsne_results["experiments_string"] == experiment
            ]
            umap_1_values = local_results["tsne1"]
            umap_2_values = local_results["tsne2"]

            self.ax.scatter(
                umap_1_values,
                umap_2_values,
                c=local_results["binding"],
                marker=markers[index],
                s=local_results["size"],
                alpha=0.5,
                cmap=prefered_cmap,
                label=experiment,
            )

            x_cor = local_results["tsne1"].tolist()
            y_cor = local_results["tsne1"].tolist()
            if toxin_names == True:
                for i, txt in enumerate(list(local_results["highest_binder"])):
                    if list(local_results["kds"])[i] > 0:
                        self.ax.annotate(txt, (x_cor[i], y_cor[i]))
            else:
                pass

        self.add_colorbar(colorbar_settings, "Binding", sm)


# sequencing_report_path = r"C:\Users\nilsh\my_projects\ExpoSeq\tmp_test\test_report.csv"
# sequencing_report = pd.read_csv(sequencing_report_path)
# peptides, selected_rows, tsne_results = PrepareData().tidy(sequencing_report, ["GeneMind_TRABkit_DNA77_300ng_repl1_L01_R1_001", "GeneMind_TRABkit_DNA80_300ng_repl1_L01_R1_001"], region_of_interest = "aaSeqCDR3", batch_size = 100)
# print(peptides)
# print(selected_rows)
# print(tsne_results)
